import re
import sys
import os

input_file = sys.argv[1] # "D:\GitHub\SquidJump\squid.list"
label_file = sys.argv[2]  # "D:\GitHub\SquidJump\squid.tass"
output_file = sys.argv[3]  # "D:\GitHub\SquidJump\squid.pdb"
root = os.getcwd()
print("input is " + input_file)
print("root is " + root)

file_lut = {}
line_to_addresstoline = {}
label_lut = {}

# line:file address label
isLabel = re.compile(r"^([0-9\:]+)\s\.([0-9a-f]{4})[\t ]*([a-zA-Z0-9]+[a-zA-Z0-9_]*)$")
# line:file address PC? hexcode raw full
splitCode = re.compile(r"^([0-9\:]+)\s\.([0-9a-f]{4})\t([0-9a-f]{4})?\s*(([0-9a-f][0-9a-f][ \t]){1,3})\t*([a-zA-Z0-9$#\(\) \,]+)\s*[+\-\w_]*?\s*([\w \$%#<>\+=\-\*\/\,\.\^\(\)_;&\s]+)")

class AddressToLine(object):
    def __init__(self):
        self.address = ""
        self.file = 0
        self.line = 0


class LabelData(object):
    def __init__(self):
        self.address = ""
        self.name = ""
        self.used = 0
        self.addr_flag = 0


def get_line_file(part, current_file):
    if ":" in part:
        line_num_parts = part.split(":")
        return int(line_num_parts[0]), int(line_num_parts[1])
    else:
        return int(part), current_file


def add_addr_line(matched, start_file_num):
    (line_num, file_num) = get_line_file(matched[1], start_file_num)
    local_addr = matched[3]  # we want PC
    if not local_addr:  # which might not matter
        local_addr = matched[2]  # so use offset
    if local_addr not in line_to_addresstoline:
        addr_line = AddressToLine()
        addr_line.address = local_addr
        addr_line.file = file_num
        addr_line.line = line_num
        line_to_addresstoline[local_addr] = addr_line
    return file_num


# this version if it matched as a label
def add_label_addr_line(matched, start_file_num):
    (line_num, file_num) = get_line_file(matched[1], start_file_num)
    local_addr = matched[2]  # so use offset
    if local_addr not in line_to_addresstoline:
        addr_line = AddressToLine()
        addr_line.address = local_addr
        addr_line.file = file_num
        addr_line.line = line_num
        line_to_addresstoline[local_addr] = addr_line
    return file_num


with open(input_file, "r") as tass:
    lines = tass.readlines()
    current_file_num = 1
    for line in lines:
        if line.startswith(':'):
            # then it is a file line
            parts = line.split('\t')
            number = parts[0][1:]
            file_parts = parts[1].split('file:')
            file = file_parts[1].strip()
            current_file_num = int(number)
            if number not in file_lut:
                print( os.path.isabs(file))
                if not os.path.isabs(file):
                    file = os.path.join(root, file)
                file_lut[number] = file
                print("found ", number, " which is file ", file)
        else:
            match = isLabel.match(line)
            if match:
                current_file_num = add_label_addr_line(match, current_file_num)
            else:
                match = splitCode.match(line)
                if match:
                    current_file_num = add_addr_line(match, current_file_num)

with open(label_file, "r") as tass_labels:
    lines = tass_labels.readlines()
    for line in lines:
        last_colon = line.rfind(":")
        if last_colon:
            if "bytes(" not in line:
                assign = line[last_colon+1:]
                parts = assign.split("=")
                name = parts[0].strip()
                if len(name) > 0:
                    value = parts[1].strip()
                    ld = LabelData()
                    ld.name = name
                    if value.startswith("$"):
                        # we have hex
                        ld.address = value[1:]
                        ld.addr_flag = 0
                    elif value.startswith('"'):
                        continue  # we don't want to keep it
                    elif value.startswith('('):
                        continue
                    elif value.startswith('['):
                        continue
                    else:
                        if '.' in value:
                            # a float
                            v = float(value)
                            ld.address = "%04X" % int(v)
                            ld.addr_flag = 1
                        if '#' in value:
                            continue # we don't want it as a label as its a constant
                        else:
                            # we have constant, mostly
                            ld.address = "%04X" % int(value)
                            ld.addr_flag = 1
                    label_lut[name] = ld

with open(output_file, "w") as pdb:
    pdb.write("FILES:"+str(len(file_lut))+"\n")
    for key in file_lut:
        int_key = int(key) - 1
        pdb.write(str(int_key)+":"+file_lut[key]+"\n")
    pdb.write("ADDRS:"+str(len(line_to_addresstoline))+"\n")
    for key in line_to_addresstoline:
        struct = line_to_addresstoline[key]
        out_line = "$%s:0:%d:%d\n" % (struct.address, struct.file-1, struct.line)
        pdb.write(out_line)
    pdb.write("LABELS:"+str(len(label_lut))+"\n")
    for key in label_lut:
        struct = label_lut[key]
        if len(struct.name) > 0:
            out_line = "$%s:0:%s:1:%d\n" % (struct.address, struct.name, struct.addr_flag)
        pdb.write(out_line)

